# SPDX-License-Identifier: Apache-2.0
# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2024

import torch
from torch import Tensor
from tqdm import tqdm
import requests
import gc
from PIL import Image

from torch.nn.attention import sdpa_kernel, SDPBackend

from ..core.quantize import HQQLinear, HQQBackend, BaseQuantizeConfig
from ..models.hf.base import AutoHQQHFModel
from .patching import patch_hqq_to_aoint4, prepare_for_inference
from .generation_hf import patch_accelerate_device_hook


class HQQGroupedGemm(torch.nn.Module):
    def __init__(
        self,
        grouped_gemm_layer,
        num_active_experts,
        quant_config,
        compute_dtype=torch.bfloat16,
        device="cuda:0",
    ):
        super().__init__()
        self.quant_config = quant_config
        self.num_active_experts = num_active_experts
        self.compute_dtype = compute_dtype
        self.device = device
        self.quant_expert(grouped_gemm_layer)

    def quant_expert(self, mlp_layer, backend="torchao_int4"):
        weight = mlp_layer.weight
        num_experts, in_features, out_features = weight.shape

        weight_int4pack, scales_and_zeros = [], []
        for j in range(num_experts):
            hqq_layer = patch_hqq_to_aoint4(
                HQQLinear.from_weights(
                    weight=weight[j].T,
                    bias=None,
                    quant_config=self.quant_config,
                    compute_dtype=self.compute_dtype,
                    device=self.device,
                    del_orig=True,
                ),
                None,
            )
            weight_int4pack.append(hqq_layer.weight_int4pack[None, :])
            scales_and_zeros.append(hqq_layer.scales_and_zeros[None, :])

        self.weights = torch.cat(weight_int4pack)
        self.scales_and_zeros = torch.cat(scales_and_zeros)
        self.in_features = in_features
        self.out_features = out_features
        self.num_experts = num_experts
        self.group_size = self.quant_config["weight_quant_params"]["group_size"]

        del hqq_layer
        torch.cuda.empty_cache()

    def forward(self, input, tokens_per_expert):
        num_tokens = input.shape[0]
        forward_fct = (
            self.forward_seq_gemm_prefill
            if (num_tokens > self.num_active_experts)
            else self.forward_seq_gemm_decode
        )
        return forward_fct(
            input,
            tokens_per_expert,
            self.num_experts,
            self.out_features,
            self.weights,
            self.scales_and_zeros,
            self.group_size,
        )

    def forward_seq_gemm_prefill(
        self,
        input: Tensor,
        tokens_per_expert: Tensor,
        num_experts: int,
        out_features: int,
        weights: Tensor,
        meta_data: Tensor,
        group_size: int,
    ) -> Tensor:
        num_tokens = input.shape[0]
        output = torch.empty(
            num_tokens, out_features, dtype=input.dtype, device=input.device
        )

        cumsum_num_tokens = torch.zeros(
            len(tokens_per_expert) + 1, dtype=torch.long, device=input.device
        )
        cumsum_num_tokens[1:] = torch.cumsum(tokens_per_expert, dim=0)

        num_active_experts = (
            (tokens_per_expert > 0).sum().int()
        )  # this should be static for torch.compile
        selected_experts = torch.topk(
            torch.diff(cumsum_num_tokens), k=num_active_experts, dim=0, largest=True
        )[1].sort()[0]

        _weights = weights[selected_experts]
        _scales = meta_data[selected_experts]
        _start = cumsum_num_tokens[selected_experts]
        _end = cumsum_num_tokens[selected_experts + 1]

        for i in range(num_active_experts):
            _s, _e = _start[i], _end[i]
            output[_s:_e] = self.matmul(
                input[_s:_e], _weights[i], _scales[i], group_size, out_features
            )

        return output

    def forward_seq_gemm_decode(
        self,
        input: Tensor,
        tokens_per_expert: Tensor,
        num_experts: int,
        out_features: int,
        weights: Tensor,
        meta_data: Tensor,
        group_size: int,
    ) -> Tensor:
        num_tokens = input.shape[0]
        output = torch.empty(
            num_tokens, out_features, dtype=input.dtype, device=input.device
        )

        cumsum_num_tokens = torch.zeros(
            len(tokens_per_expert) + 1, dtype=torch.long, device=input.device
        )
        cumsum_num_tokens[1:] = torch.cumsum(tokens_per_expert, dim=0)

        num_active_experts = self.num_active_experts
        selected_experts = torch.topk(
            torch.diff(cumsum_num_tokens), k=num_active_experts, dim=0, largest=True
        )[1].sort()[0]

        _weights = weights[selected_experts]
        _scales = meta_data[selected_experts]
        _start = cumsum_num_tokens[selected_experts]
        _end = cumsum_num_tokens[selected_experts + 1]

        for i in range(num_active_experts):
            _s, _e = i, i + 1
            output[_s:_e] = self.matmul(
                input[_s:_e], _weights[i], _scales[i], group_size, out_features
            )

        return output

    def matmul(
        self,
        x: Tensor,
        weight_int4pack: Tensor,
        scales_and_zeros: Tensor,
        groupsize: int,
        out_features: int,
    ) -> Tensor:
        origin_x_size = x.size()
        x = x.reshape(-1, origin_x_size[-1])
        c = torch.ops.aten._weight_int4pack_mm(
            x, weight_int4pack, groupsize, scales_and_zeros
        )
        new_shape = origin_x_size[:-1] + (out_features,)
        c = c.reshape(new_shape)
        return c


def quantize_model(
    model,
    num_active_experts=6,
    attn_quant_config=BaseQuantizeConfig(nbits=4, group_size=64, axis=1),
    experts_quant_config=BaseQuantizeConfig(nbits=4, group_size=64, axis=1),
    compute_dtype=torch.bfloat16,
    device="cuda:0",
):
    for i in tqdm(range(len(model.language_model.model.layers))):
        model.language_model.model.layers[i].mlp.experts.fc1 = HQQGroupedGemm(
            model.language_model.model.layers[i].mlp.experts.fc1,
            num_active_experts=num_active_experts,
            quant_config=experts_quant_config,
            compute_dtype=compute_dtype,
            device=device,
        )
        model.language_model.model.layers[i].mlp.experts.fc2 = HQQGroupedGemm(
            model.language_model.model.layers[i].mlp.experts.fc2,
            num_active_experts=num_active_experts,
            quant_config=experts_quant_config,
            compute_dtype=compute_dtype,
            device=device,
        )
    gc.collect()

    # Quantize the rest
    AutoHQQHFModel.quantize_model(
        model.language_model,
        quant_config=attn_quant_config,
        compute_dtype=compute_dtype,
        device=device,
    )

    # Remove losses
    # import moe_llm
    # moe_llm.apply_z_loss   = lambda logits: logits
    # moe_llm.apply_aux_loss = lambda logits, tokens_per_expert, scores: scores

    # Move the vision model to the device
    model.multi_modal_projector = model.multi_modal_projector.to(device)
    model.vision_tower = model.vision_tower.to(device)

    # Optimize
    HQQLinear.set_backend(
        HQQBackend.ATEN
        if experts_quant_config["weight_quant_params"]["axis"] == 0
        else HQQBackend.PYTORCH_COMPILE
    )
    prepare_for_inference(model.language_model, backend="torchao_int4", verbose=True)


def generate(model, processor, img_path, prompt, max_new_tokens=500, do_sample=True):
    image = Image.open(requests.get(img_path, stream=True).raw)

    messages = [
        {
            "role": "user",
            "content": [
                {"text": None, "type": "image"},
                {"text": prompt, "type": "text"},
            ],
        }
    ]

    text = processor.apply_chat_template(messages, add_generation_prompt=True)
    inputs = processor(text=text, images=image, return_tensors="pt")
    inputs["pixel_values"] = inputs["pixel_values"].to(model.dtype)
    inputs = {k: v.to(model.device) for k, v in inputs.items()}

    with torch.inference_mode():
        output = model.generate(
            **inputs,
            max_new_tokens=max_new_tokens,
            stop_strings=["<|im_end|>"],
            tokenizer=processor.tokenizer,
            do_sample=do_sample,
            temperature=0.9,
            cache_implementation="static",
        )
        output_ids = output[0][inputs["input_ids"].shape[1] :]
        result = processor.decode(output_ids, skip_special_tokens=True)

    return result


def patch_model_for_compiled_runtime(
    model, processor, warmup=True, patch_accelerate=True
):
    if patch_accelerate:
        patch_accelerate_device_hook()

    torch._dynamo.config.inline_inbuilt_nn_modules = False  # torch 2.5.0 fix

    model.language_model.config.use_cache = True
    model.language_model.generation_config.cache_implementation = "static"
    model.eval()

    forward_compiled = torch.compile(
        model.language_model.forward, mode="reduce-overhead", fullgraph=True
    )
    forward_simple = model.language_model.forward

    if processor.tokenizer.pad_token is None:
        processor.tokenizer.add_special_tokens({"pad_token": "<<[PAD]>>"})

    def custom_forward(*args, **kwargs):
        # Decoding phase
        if (
            (len(args) > 0 and args[0].shape[-1] == 1)
            or ("input_ids" in kwargs and kwargs["input_ids"].shape[-1] == 1)
            or ("inputs_embeds" in kwargs and kwargs["inputs_embeds"].shape[1] == 1)
        ):
            out_fct = forward_compiled

            with sdpa_kernel([SDPBackend.MATH]):
                out = out_fct(*args, **kwargs)

            return out

        # Prefill phase
        out_fct = forward_simple
        out = out_fct(*args, **kwargs)
        return out

    model.language_model.forward = custom_forward

    if warmup:
        for _ in tqdm(range(5)):
            generate(
                model,
                processor,
                img_path="https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/diffusers/cat.png",
                prompt="what is the image?",
            )



# SPDX-License-Identifier: Apache-2.0
# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2023

import torch
from torch import float16, bfloat16, float32
from torch import Tensor, nn
import numpy as np
from .quantize import HQQLinear, Quantizer
from .utils import cleanup
from typing import Union


# Return trainable weight matrix
def _get_dense_param(
    in_features: int,
    out_features: int,
    device="cuda",
    trainable: bool = True,
    dtype: torch.dtype = float32,
) -> Tensor:
    W = (
        nn.Linear(in_features, out_features, bias=None)
        .weight.data.t()
        .to(dtype)
        .to(device)
        .contiguous()
    )
    return nn.Parameter(W, requires_grad=trainable)


# Applies LoRA to a linear module (include HQQLinear)
class HQQLinearLoRA(nn.Module):
    def __init__(self, linear_layer: nn.Module, peft_config: dict):
        super().__init__()

        # Device
        self.device = (
            linear_layer.device
            if hasattr(linear_layer, "device")
            else next(linear_layer.parameters()).device
        )
        self.train_dtype = (
            peft_config["train_dtype"]
            if ("train_dtype" in peft_config)
            else torch.float
        )

        # Linear layer
        self.linear_layer = linear_layer
        self.in_features = linear_layer.in_features
        self.out_features = linear_layer.out_features

        # Bias
        self.bias = None if (linear_layer.bias is None) else linear_layer.bias.clone()
        self.linear_layer.bias = None
        self.train_bias = (
            peft_config["train_bias"] if ("train_bias" in peft_config) else False
        )

        if self.bias is not None:
            self.bias = nn.Parameter(self.bias, requires_grad=self.train_bias)
            if self.train_bias:
                self.bias.data = self.bias.data.to(self.train_dtype)

        if (self.bias is None) and self.train_bias:
            self.bias = nn.Parameter(
                torch.zeros(
                    (self.out_features,), device=self.device, dtype=self.train_dtype
                ),
                requires_grad=True,
            )

        # Dropout
        if "dropout" in peft_config:
            self.peft_drop = (
                nn.Dropout(p=peft_config["dropout"])
                if (peft_config["dropout"] > 0.0)
                else nn.Identity()
            )
        else:
            self.peft_drop = nn.Identity()

        # LoRA A/B
        self.peft_config = peft_config
        self.lora_alpha = peft_config["lora_alpha"]
        self.r = peft_config["r"]
        self.scaling = self.lora_alpha / self.r

        self.lora_A = _get_dense_param(
            self.in_features,
            self.r,
            device=self.device,
            trainable=True,
            dtype=self.train_dtype,
        )

        self.lora_B = _get_dense_param(
            self.r,
            self.out_features,
            device=self.device,
            trainable=True,
            dtype=self.train_dtype,
        )

        # LoRA weights init
        if "lora_init" in peft_config:
            # Set lora init
            assert (
                peft_config["lora_init"]["lora_A"].shape[0],
                peft_config["lora_init"]["lora_B"].shape[1],
            ) == (self.in_features, self.out_features), (
                "Invalid init LoRA weight shapes. Expected: lora_A: "
                + str(self.in_features)
                + " x r , lora_B: r x "
                + str(self.out_features)
                + ")"
            )
            self.lora_A.data = peft_config["lora_init"]["lora_A"].to(
                device=self.device, dtype=self.train_dtype
            )
            self.lora_B.data = peft_config["lora_init"]["lora_B"].to(
                device=self.device, dtype=self.train_dtype
            )
        else:
            # Init weights, as as the original LoRA implementation
            nn.init.kaiming_uniform_(self.lora_A, a=np.sqrt(5))
            nn.init.zeros_(self.lora_B)


        #Compute dtype = used for inference
        if(hasattr(self.linear_layer, 'compute_dtype')): #HQQ layer
            self.compute_dtype = self.linear_layer.compute_dtype
        else: #Regular layer
            fp_param = [p for p in layer.parameters() if p.is_floating_point()]
            if(len(fp_param)>0):
                self.compute_dtype = fp_param[0].dtype
            else:
                self.compute_dtype = self.train_dtype


    def forward_lora(self, x: Tensor) -> Tensor:  # output is self.train_dtype
        return (
            torch.matmul(
                torch.matmul(self.peft_drop(x.to(self.lora_A.dtype)), self.lora_A),
                self.lora_B,
            )
            * self.scaling
        )

    def forward(self, x: Tensor) -> Tensor:
        x_dtype = x.dtype

        # Forward with the base linear layer
        out = self.linear_layer(x)

        # LoRA
        if self.train_bias:
            out += (self.forward_lora(x) + self.bias).to(x_dtype)

        else:
            out += self.forward_lora(x).to(x_dtype)
            if self.bias is not None:
                out += self.bias

        return out

    def merge_and_quantize(self, quant_config: dict) -> nn.Module:
        # not easy to figure out the type for "any" potential linear layer (nn.Linear, HQQLinear, etc.)
        for _dtype in [float16, bfloat16, float32]:
            try:
                # Get initial weights
                W = self.linear_layer(
                    torch.eye(self.in_features, device=self.device, dtype=_dtype)
                ).t()  # == self.linear_layer.dequantize()
            except Exception:
                pass

        # Merge weights
        W += (
            (torch.matmul(self.lora_A.data, self.lora_B.data) * self.scaling)
            .t()
            .to(W.dtype)
        )

        # New HQQ layer
        new_hqq_layer = HQQLinear(None, quant_config)
        new_hqq_layer.bias = None if (self.bias is None) else self.bias.clone()
        new_hqq_layer.quantize(W, **quant_config)

        return new_hqq_layer

    def cast(self, dtype: torch.dtype = float16):
        self.lora_A.data = self.lora_A.data.to(dtype)
        self.lora_B.data = self.lora_B.data.to(dtype)
        if self.bias is not None:
            self.bias.data = self.bias.data.to(dtype)
        if isinstance(self.scaling, nn.Parameter):
            self.scaling.data = self.scaling.data.to(dtype)
        if isinstance(self.scaling, Tensor):
            self.scaling = self.scaling.to(dtype)
        return self

    def state_dict(self):
        return {
            "lora_A": self.lora_A.data,
            "lora_B": self.lora_B.data,
            "scaling": self.scaling,
            "bias": self.bias,
        }

    def load_state_dict(self, state_dict):
        self.lora_A.data = state_dict["lora_A"].data.to(device=self.device, dtype=self.compute_dtype)
        self.lora_B.data = state_dict["lora_B"].data.to(device=self.device, dtype=self.compute_dtype)

        if state_dict["bias"] is not None:
            self.bias.data = state_dict["bias"].data.to(device=self.device, dtype=self.compute_dtype)

        # Handle different use-cases of scaling
        if isinstance(state_dict["scaling"], (int, float)):
            self.scaling = state_dict["scaling"]
        if isinstance(state_dict["scaling"], nn.Parameter):
            self.scaling.data = state_dict["scaling"].data.to(device=self.device, dtype=self.compute_dtype)
        if isinstance(state_dict["scaling"], Tensor):
            self.scaling = state_dict["scaling"].to(device=self.device, dtype=self.compute_dtype)


# LoRA with fake quantization
class HQQLinearLoRAWithFakeQuant(HQQLinearLoRA):
    def __init__(self, linear_layer: nn.Module, peft_config: dict):
        super(HQQLinearLoRAWithFakeQuant, self).__init__(linear_layer, peft_config)
        self.quant_param = peft_config["quant_param"]

    # @torch.no_grad()
    # @torch.compile()
    def fake_quant(self, weight):
        if self.quant_param:
            W_q, meta = Quantizer.quantize(
                weight, **self.quant_param, bitpack=False
            )  # todo: clone() tensor
            weight_est = Quantizer.dequantize(W_q, meta)
        else:
            weight_est = weight
        return weight_est

    def forward(self, x: Tensor) -> Tensor:
        x_dtype = x.dtype

        # Get initial weights
        W = self.linear_layer(
            torch.eye(self.in_features, device=self.device, dtype=x_dtype)
        ).t()  # == self.linear_layer.dequantize()

        # Merge weights
        W += (torch.matmul(self.lora_A, self.lora_B) * self.scaling).t().to(W.dtype)

        # Fake quant
        W = self.fake_quant(W).to(x_dtype)

        # Matmul
        out = torch.matmul(x, W.t())

        # Bias
        if self.bias is not None:
            out += self.bias

        out = out.to(x_dtype)

        return out


# Experimental
class HQQLinearGroupedProj(nn.Module):
    def __init__(self, linear_layer: nn.Module, peft_config: dict):
        super().__init__()

        # Device
        self.device = (
            linear_layer.device
            if hasattr(linear_layer, "device")
            else next(linear_layer.parameters()).device
        )
        self.train_dtype = (
            peft_config["train_dtype"]
            if ("train_dtype" in peft_config)
            else torch.float
        )

        # Linear layer
        self.linear_layer = linear_layer
        self.in_features = linear_layer.in_features
        self.out_features = linear_layer.out_features
        self.bias = None if (linear_layer.bias is None) else linear_layer.bias.clone()

        # Turn-off bias in the linear layer
        self.linear_layer.bias = None

        # Group proj
        self.peft_config = peft_config
        self.proj_size = peft_config["proj_size"]
        self.proj_num = peft_config["proj_num"]
        self.proj = nn.Parameter(
            torch.stack(
                [torch.eye(self.proj_size, dtype=self.train_dtype, device=self.device)]
                * self.proj_num
            )
        )
        if peft_config["zero_trainable"]:
            self.linear_layer.meta["zero"] = nn.Parameter(
                self.linear_layer.meta["zero"].to(self.train_dtype), requires_grad=True
            )

    @torch.compile()
    def forward(self, x: Tensor) -> Tensor:
        x_dtype = x.dtype

        # Forward with base linear
        with torch.no_grad():
            W = self.linear_layer.dequantize().clone()
        # W = self.linear_layer(torch.eye(self.in_features, device=self.device, dtype=x_dtype)).t()
        shape = W.shape

        # Grouped proj
        proj_b, gs = self.proj.shape[0], self.proj.shape[1]
        W = (
            torch.matmul(self.proj, W.reshape((proj_b, gs, -1)).to(self.proj.dtype))
            .to(x_dtype)
            .reshape(shape)
        )

        # Matmul
        out = torch.matmul(x, W.t())

        # Bias
        if self.bias is not None:
            out += self.bias

        out = out.to(x_dtype)

        return out

    def cast(self, dtype: torch.dtype = float16):
        self.proj.data = self.proj.data.to(dtype)
        self.linear_layer.meta["zero"] = self.linear_layer.meta["zero"].to(dtype)
        if self.bias is not None:
            if self.bias.requires_grad:
                self.bias.data = self.bias.data.to(dtype)
            else:
                self.bias = self.bias.to(dtype)
        return self

    def state_dict(self):
        return {
            "proj": self.proj.data,
            "bias": self.bias,
            "peft_config": self.peft_config,
        }

    def load_state_dict(self, state_dict):
        self.proj.data = state_dict["proj"].data.to(self.device)
        self.bias = state_dict["bias"] if ("bias" in state_dict) else None
        self.bias = self.bias.to(self.device) if (self.bias is not None) else None
        self.peft_config = state_dict["peft_config"]


_HQQ_LORA_CLASSES = [HQQLinearLoRA, HQQLinearLoRAWithFakeQuant, HQQLinearGroupedProj]
_HQQ_LORA_MAPPING = {
    "default": HQQLinearLoRA,
    "lora_with_fakequant": HQQLinearLoRAWithFakeQuant,
    "grouped_proj": HQQLinearGroupedProj,
}


def is_hqq_lora_layer(layer):
    return type(layer) in _HQQ_LORA_CLASSES


##################################################################################################################
def autoname_modules(model):
    for name, module in model.named_modules():
        module.name = name


# Patching functions
def patch_linear_add_peft(layer: nn.Module, patch_params: Union[dict,None]) -> nn.Module:
    _peft_config = patch_params
    if _peft_config:
        lora_type = (
            _peft_config["lora_type"] if ("lora_type" in _peft_config) else "default"
        )
        new_layer = _HQQ_LORA_MAPPING[lora_type](layer, _peft_config)
    else:
        new_layer = layer
    return new_layer


def patch_linear_merge_peft(layer: nn.Module, patch_params: Union[dict,None]) -> nn.Module:
    _quant_config = patch_params
    if _quant_config:
        new_layer = layer.merge_and_quantize(_quant_config)
        del layer
        cleanup()
    else:
        new_layer = layer
    return new_layer


def patch_linear_cast_peft(layer: nn.Module, patch_params: Union[dict,None]) -> nn.Module:
    if is_hqq_lora_layer(layer):
        layer.cast(patch_params)
    return layer


# Putting it all together
class PeftUtils:
    @classmethod
    def get_base_class(cls, model, base_class):
        # Get base class
        if base_class is None:
            base_class = getattr(model, "base_class", None)

        assert (
            base_class is not None
        ), "You need to provide the base HQQ class (LlamaHQQ, MixtralHQQ, etc.) as model.base_class or as an argument base_class=LlamaHQQ. Use AutoHQQHFMode if the model architecture is not supported."

        return base_class

    @classmethod
    def add_lora(
        cls, model, peft_config, base_class: bool = None, verbose: bool = True
    ) -> None:
        # Base classs
        base_class = cls.get_base_class(model, base_class)

        base_class.setup_model(model)

        # Freeze
        for param in model.parameters():
            param.requires_grad = False

        # Patch
        base_class.patch_linearlayers(
            model, patch_linear_add_peft, peft_config, verbose=verbose
        )

        # Rename modules
        autoname_modules(model)

        # add config to model
        model.peft_config = peft_config

    @classmethod
    def merge_lora(
        cls, model, merge_lora_params, base_class: bool = None, verbose: bool = True
    ) -> None:
        # Base classs
        base_class = cls.get_base_class(model, base_class)

        # Patch
        base_class.patch_linearlayers(
            model, patch_linear_merge_peft, merge_lora_params, verbose=verbose
        )

    @classmethod
    def cast_lora_weights(
        cls, model, dtype: torch.dtype, base_class: bool = None, verbose: bool = True
    ) -> None:
        # Base classs
        base_class = cls.get_base_class(model, base_class)

        base_class.setup_model(model)

        # Linear tags
        linear_tags = model.linear_tags

        # Patch
        base_class.patch_linearlayers(
            model,
            patch_linear_cast_peft,
            dict([(linear_tag, dtype) for linear_tag in linear_tags]),
            verbose=verbose,
        )

    @classmethod
    def save_lora_weights(
        cls, model, filename: str, base_class: bool = None, verbose: bool = True
    ) -> None:
        # Base classs
        base_class = cls.get_base_class(model, base_class)

        base_class.setup_model(model)

        lora_global_params = {}

        def _patch_linear_save_weights(layer, patch_params, return_layer=True):
            if is_hqq_lora_layer(layer):
                lora_global_params[layer.name] = layer.state_dict()
            if return_layer:
                return layer

        # Linear tags
        linear_tags = model.linear_tags

        # Patch
        base_class.patch_linearlayers(
            model,
            _patch_linear_save_weights,
            dict([(linear_tag, None) for linear_tag in linear_tags]),
            verbose=verbose,
        )

        # save
        torch.save(
            {"peft_config": model.peft_config, "parameters": lora_global_params},
            filename,
        )

    @classmethod
    def load_lora_weights(
        cls, model, filename: str, base_class: bool = None, verbose: bool = True
    ) -> None:
        # Base classs
        base_class = cls.get_base_class(model, base_class)

        base_class.setup_model(model)

        lora_data = torch.load(filename, map_location="cpu", weights_only=True)

        # V0.2 format: automatically creates lora modules if the model doesn't contain it
        if ("peft_config" in lora_data) and ("parameters" in lora_data):
            peft_config = lora_data["peft_config"]
            lora_global_params = lora_data["parameters"]
            if not hasattr(model, "peft_config"):
                cls.add_lora(model=model, peft_config=peft_config)

        # v0.1 format
        else:
            if not hasattr(model, "peft_config"):
                raise Exception(
                    "Using older version of lora weights. LoRa modules should be manually added in this case."
                )
            lora_global_params = lora_data

        #Load weights / cast to compute_dtype
        def _load_weights(layer, lora_global_params):
            layer.load_state_dict(lora_global_params[layer.name])
            return layer

        def _patch_linear_load_weights(model, lora_global_params):
            for name, layer in model.named_children():
                if is_hqq_lora_layer(layer):
                    setattr(model, name, _load_weights(layer, lora_global_params))
                else:
                    _patch_linear_load_weights(layer, lora_global_params)

        _patch_linear_load_weights(model, lora_global_params)


# SPDX-License-Identifier: Apache-2.0
# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2023

import transformers
from accelerate import init_empty_weights
from ..base import BaseHQQModel, BasePatch

class BaseHQQHFModel(BaseHQQModel):
    # Save model architecture
    @classmethod
    def cache_model(cls, model, save_dir):
        # Update model architecture in the config
        model.config.architectures = [model.__class__.__name__]
        # Save config
        model.config.save_pretrained(save_dir)

    # Create empty model from config
    @classmethod
    def create_model(cls, save_dir, kwargs):
        model_kwargs = {}
        for key in ["attn_implementation"]:
            if key in kwargs:
                model_kwargs[key] = kwargs[key]

        config = transformers.AutoConfig.from_pretrained(
            cls.get_config_file(save_dir)
        )

        auto_class = transformers.AutoModel

        # Todo: add support for other auto models
        archs = config.architectures
        if len(archs) == 1:
            if ("CausalLM" in archs[0]):
                auto_class = transformers.AutoModelForCausalLM
            elif ("SequenceClassification" in archs[0]):
                auto_class = transformers.AutoModelForSequenceClassification

        with init_empty_weights():
            model = auto_class.from_config(config, **model_kwargs)

        return model


# Auto class used for HF models if no architecture was manually setup
class AutoHQQHFModel(BaseHQQHFModel, BasePatch):
    pass



# SPDX-License-Identifier: Apache-2.0
# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2024

# Rewrite generate() to support torch.compile fullgraph.
# https://gist.github.com/ArthurZucker/5dc54a3fb443e979fac437e5df7c800b

import torch
from torch.nn.attention import sdpa_kernel, SDPBackend
from typing import Union, Dict
from transformers import StaticCache
from tqdm import tqdm

WARMUP_PROMPTS = [
    "Write an essay about large language models.",
    "Tell me a funny joke!",
    "How to make a yummy chocolate cake?",
    "Who is Elon Musk?",
    "Write a Python code snippet that adds two numbers together.",
]

#The original function from accelerate breaks with torch.compile, this is a hacky fix
def patch_accelerate_device_hook():
    #https://github.com/huggingface/accelerate/blob/main/src/accelerate/utils/operations.py#L136
    import accelerate.utils.operations as ops 
    from typing import Mapping
    def _send_to_device(tensor, device,  non_blocking=False, skip_keys=None):
        if(isinstance(tensor, tuple)):
            return tuple(t.to(device) for t in tensor)
        if(isinstance(tensor, list)):
            return [t.to(device) for t in tensor]
        if(isinstance(tensor, Mapping)):
            if isinstance(skip_keys, str):
                skip_keys = [skip_keys]
            elif skip_keys is None:
                skip_keys = []
            return type(tensor)(
                {
                    k: t if k in skip_keys else _send_to_device(t, device, non_blocking=non_blocking, skip_keys=skip_keys)
                    for k, t in tensor.items()
                }
            )
        if(isinstance(tensor, (torch.Tensor, torch.nn.Parameter))):
            return tensor.to(device)
        else:
            return tensor

    ops.send_to_device = _send_to_device


# Patches a HuggingFace model to work with torch.compile + static cache
def patch_model_for_compiled_runtime(
    model, tokenizer, warmup=True, max_new_tokens=1000, patch_accelerate=True, pre_compile=None, compile_prefill=False,
):  
    if(patch_accelerate):
        patch_accelerate_device_hook()

    model.config.use_cache = True
    model.generation_config.cache_implementation = "static"
    model.eval()
    
    torch._dynamo.config.inline_inbuilt_nn_modules = False #torch 2.5.0 fix

    forward_simple = model.forward

    if(pre_compile is None):
        pre_compile = getattr(model.generation_config, 'compile_config', None) is None 

    if(pre_compile):
        if(compile_prefill):
            forward_prefill = torch.compile(model.forward, dynamic=True, fullgraph=True)
        else:
            forward_prefill = model.forward
        forward_decode = torch.compile(model.forward, mode="reduce-overhead", fullgraph=True)
    else:
        forward_prefill = model.forward
        forward_decode = model.forward

    if tokenizer.pad_token is None:
        tokenizer.add_special_tokens({"pad_token": tokenizer.eos_token})

    def custom_forward(*args, **kwargs):
        # Decoding pahse
        if (len(args) > 0 and args[0].shape[-1] == 1) or (
            "input_ids" in kwargs and kwargs["input_ids"].shape[-1] == 1
        ):
            return forward_decode(*args, **kwargs)
        else:
            #Prefill phase
            return forward_prefill(*args, **kwargs)

    model.forward = custom_forward

    # Warm-up
    if warmup:
        from tqdm import tqdm

        for prompt in tqdm(WARMUP_PROMPTS):
            model.generate(
                **tokenizer(
                    [
                        tokenizer.apply_chat_template(
                            [
                                {"role": "user", "content": prompt},
                            ],
                            tokenize=False,
                        )
                    ],
                    return_tensors="pt",
                ).to(model.device),
                max_new_tokens=max_new_tokens,
                cache_implementation="static",
                pad_token_id=tokenizer.pad_token_id,
            )


# Custom generator
class HFGenerator:
    def __init__(
        self,
        model,
        tokenizer,
        max_new_tokens: int = 1000,
        cache_size: Union[int, None] = None,
        do_sample: bool = False,
        temperature: float = 0.6,
        top_k: int = 5,
        compile: Union[str, None] = None,  # None / "partial" / "full"
        compile_options: Dict = {"mode": "reduce-overhead", "fullgraph": True},
        patch_accelerate: bool = True,
    ):
        super().__init__()

        torch._dynamo.config.cache_size_limit = 64
        torch._dynamo.config.capture_scalar_outputs = True
        torch._inductor.config.fx_graph_cache = True
        try:
            torch._dynamo.config.inline_inbuilt_nn_modules = False #torch 2.5.0 fix
        except:
            pass

        if(patch_accelerate):
            patch_accelerate_device_hook()

        self.model = model
        self.tokenizer = tokenizer
        self.device = model.device
        self.do_sample = do_sample
        self.temperature = temperature if self.do_sample else None
        self.top_k = top_k if self.do_sample else None
        self.use_cache = True  # False
        self.compile_options = compile_options

        if do_sample:
            decode_one_token = self.decode_one_token_sampled
        else:
            decode_one_token = self.decode_one_token_no_sample

        # Setup cache
        self.max_new_tokens = max_new_tokens
        if cache_size is None:
            self.cache_size = self.next_multiple(self.max_new_tokens)
        else:
            self.cache_size = cache_size

        self.max_new_tokens = min(self.max_new_tokens, self.cache_size)

        self.setup_cache()

        self.is_compiled = False
        if compile == "partial":
            self.compile_partial(decode_one_token)

        if compile == "full":
            self.compile_full()

        if hasattr(self, "decode_one_token") is False:
            self.decode_one_token = decode_one_token

        self.init()  # check this: move this before setup_cache?

        ############################
        #Cuda Graph section
        self.static_input     = torch.zeros((1, 1), device=self.device, dtype=torch.int32)
        self.static_output    = torch.zeros((1, 1), device=self.device, dtype=torch.int32)
        self.cuda_graph       = None
        self.do_capture_graph = False
        ############################

    @torch.no_grad()
    def setup_cache(self):
        try:
            self.past_key_values = StaticCache(
                config=self.model.config,
                max_batch_size=1,
                max_cache_len=self.cache_size,
                device=self.model.device,
                dtype=self.model.dtype,
            )
        except:
            self.past_key_values = StaticCache(
                self.model.config, 1, self.cache_size, self.model.device, self.model.dtype
            )

    @torch.no_grad()
    def reset_cache(self):
        self.past_key_values.reset()

    # Ideally only compile this, but it creates issues with generation: https://github.com/huggingface/transformers/issues/30351
    def compile_partial(self, decode_one_token):
        self.decode_one_token = torch.compile(decode_one_token, **self.compile_options)
        self.is_compiled = True

    @torch.inference_mode()
    def compile_full(self):
        self.model.forward = torch.compile(self.model.forward, **self.compile_options)

        with sdpa_kernel([SDPBackend.MATH]):
            for ctx_size in [1] * 10:
                self.model(
                    torch.ones(
                        (1, ctx_size), dtype=torch.int32, device=self.model.device
                    ),
                    use_cache=False,
                ).logits  # check this: use_cache=True?
        self.is_compiled = True

    # torch.compile needs about 5-10 runs to warmup
    def warmup(self, max_samples=-1):
        for prompt in WARMUP_PROMPTS[:max_samples if max_samples>0 else len(WARMUP_PROMPTS)]:
            self.generate(prompt, print_tokens=False)
        return self

    def next_multiple(self, val):  # next power of 2
        vals = [2**i for i in range(5, 20)]  # [32, 64, ...]
        new_val = vals[[i for i in range(len(vals)) if (vals[i] - val) > 0][0]]
        return new_val

    def init(self):
        # Setup inference mode
        self.tokenizer.add_bos_token = False
        self.tokenizer.add_eos_token = False
        if not self.tokenizer.pad_token:
            self.tokenizer.add_special_tokens({"pad_token": "<<[PAD]>>"})
        self.tokenizer.padding_side = "right"
        self.model.eval()
        self.model.generation_config.cache_implementation = "static"
        self.model.config.use_cache = True

    # Copied from https://gist.github.com/ArthurZucker/af34221def212259b43d55a2811d2dbb
    def multinomial_sample_one_no_sync(self, probs_sort):
        q = torch.empty_like(probs_sort).exponential_(1)
        return torch.argmax(probs_sort / q, dim=-1, keepdim=True).to(dtype=torch.int)

    # Copied from https://gist.github.com/ArthurZucker/af34221def212259b43d55a2811d2dbb
    def logits_to_probs(self, logits, temperature=1.0, top_k=None):
        logits = logits / max(temperature, 1e-5)
        if top_k is not None:
            v, _ = torch.topk(logits, min(top_k, logits.size(-1)))
            pivot = v.select(-1, -1).unsqueeze(-1)
            logits = torch.where(logits < pivot, -float("Inf"), logits)
        probs = torch.nn.functional.softmax(logits, dim=-1)
        return probs

    # Copied from https://gist.github.com/ArthurZucker/af34221def212259b43d55a2811d2dbb
    def sample(self, logits, temperature, top_k):
        probs = self.logits_to_probs(logits[:, -1], temperature, top_k)
        idx_next = self.multinomial_sample_one_no_sync(probs)
        return idx_next, probs

    def decode_one_token_no_sample(
        self,
        cur_token,
        input_pos,
        cache_position,
        past_key_values,
        temperature=None,
        top_k=None,
    ):
        out = self.model(
            cur_token,
            # position_ids=input_pos,
            cache_position=cache_position,
            past_key_values=past_key_values,
            return_dict=True,
            use_cache=self.use_cache,
        )
        logits, self.past_key_values = out.logits, out.past_key_values
        new_token = torch.argmax(logits[:, -1], dim=-1)[:, None]
        return new_token

    def decode_one_token_sampled(
        self,
        cur_token,
        input_pos,
        cache_position,
        past_key_values,
        temperature=0.6,
        top_k=5,
    ):
        out = self.model(
            cur_token,
            # position_ids=input_pos,
            cache_position=cache_position,
            past_key_values=past_key_values,
            return_dict=True,
            use_cache=self.use_cache,
        )
        logits, self.past_key_values = out.logits, out.past_key_values
        new_token = self.sample(logits, temperature=temperature, top_k=top_k)[0]
        return new_token

    # Setup cache and variables
    def setup(self, inputs, max_new_tokens):
        self.reset_cache()
        self.inputs = inputs
        self.batch_size, self.seq_length = self.inputs["input_ids"].shape
        self.cache_position = torch.arange(self.seq_length, device=self.device)
        self.generated_ids = torch.zeros(
            self.batch_size,
            self.seq_length + max_new_tokens + 1,
            dtype=torch.int,
            device=self.device,
        )
        self.generated_ids[:, self.cache_position] = self.inputs["input_ids"].to(
            torch.int
        )

    # Pre-fill phase
    def prefill(self):
        out = self.model(
            **self.inputs,
            cache_position=self.cache_position,
            past_key_values=self.past_key_values,
            return_dict=True,
            use_cache=self.use_cache,
        )
        logits, self.past_key_values = out.logits, out.past_key_values
        next_token = torch.argmax(logits[:, -1], dim=-1)[:, None]
        self.generated_ids[:, self.seq_length] = next_token[:, 0]
        self.cache_position = torch.tensor([self.seq_length], device=self.device, dtype=torch.long)
        self.begin_gen_position = self.cache_position.item()
        return next_token

    # generate one token at a time
    def gen_next_token_raw(self, next_token):
        with sdpa_kernel([SDPBackend.MATH]):
            next_token = self.decode_one_token(
                next_token.clone(),
                None,
                cache_position=self.cache_position + 1,
                past_key_values=self.past_key_values,
                temperature=self.temperature,
                top_k=self.top_k,
            )
        self.cache_position += 1
        self.generated_ids[:, self.cache_position] = next_token.int()
        return next_token

    def gen_next_token(self, next_token):
        return self.gen_next_token_raw(next_token)

    def enable_cuda_graph(self):
        #Warm-up
        self.warmup(1)

        #Enable 
        self.gen_next_token = self.gen_next_token_withgraph_v1
        self.do_capture_graph = True

        #Capture
        self.warmup(1)

        return self

    def gen_next_token_withgraph_v1(self, next_token):
        self.static_input.copy_(next_token)
        self.cache_position += 1

        if(self.do_capture_graph):
            self.stream = torch.cuda.Stream()
            torch.cuda.synchronize()

            #Warm-up
            with torch.cuda.stream(self.stream):
                for _ in range(3):
                    with sdpa_kernel([SDPBackend.MATH]):
                        _ = self.decode_one_token(
                            self.static_input,
                            None,
                            cache_position=self.cache_position,
                            past_key_values=self.past_key_values,
                            temperature=self.temperature,
                            top_k=self.top_k,
                        )
                torch.cuda.synchronize()
  
            #Capture
            self.cuda_graph = torch.cuda.CUDAGraph()
            with torch.cuda.stream(self.stream):
                self.cuda_graph.capture_begin()
                with sdpa_kernel([SDPBackend.MATH]):
                    out = self.decode_one_token(
                        self.static_input,
                        None,
                        cache_position=self.cache_position,
                        past_key_values=self.past_key_values,
                        temperature=self.temperature,
                        top_k=self.top_k,
                    )
                    self.static_output.copy_(out)
                self.cuda_graph.capture_end()
            torch.cuda.synchronize()
            
            #Turn-off
            self.do_capture_graph = False
        else:
            self.cuda_graph.replay()

        next_token = self.static_output
        self.generated_ids[:, self.cache_position] = next_token.int()
        return next_token

    def gen_next_token_withgraph_v2(self, next_token):
        if(self.do_capture_graph):
            self.static_input.copy_(next_token)
            self.stream = torch.cuda.Stream()
            torch.cuda.synchronize()

            #Warm-up
            with torch.cuda.stream(self.stream):
                for _ in range(3):
                    with sdpa_kernel([SDPBackend.MATH]):
                        _ = self.decode_one_token(
                            self.static_input,
                            None,
                            cache_position=self.cache_position + 1,
                            past_key_values=self.past_key_values,
                            temperature=self.temperature,
                            top_k=self.top_k,
                        )
                torch.cuda.synchronize()
  
            #Capture
            self.cuda_graph = torch.cuda.CUDAGraph()
            with torch.cuda.stream(self.stream):
                self.cuda_graph.capture_begin()
                with sdpa_kernel([SDPBackend.MATH]):
                    out = self.decode_one_token(
                        self.static_input,
                        None,
                        cache_position=self.cache_position + 1,
                        past_key_values=self.past_key_values,
                        temperature=self.temperature,
                        top_k=self.top_k,
                    )
                    self.static_output.copy_(out)
                self.cuda_graph.capture_end()
            torch.cuda.synchronize()
            
            #Turn-off
            self.do_capture_graph = False
        else:
            self.static_input.copy_(next_token)
            self.cuda_graph.replay()

        next_token = self.static_output
        self.cache_position += 1
        self.generated_ids[:, self.cache_position] = next_token.int()
        return next_token

    def print_current_token(self, output_text_len):
        output_text = self.tokenizer.decode(self.generated_ids[0, self.begin_gen_position : self.cache_position + 1])
        printable_text = output_text[output_text_len:]
        output_text_len = len(output_text)
        print(printable_text, end="", flush=True)
        return output_text_len

    def next_token_iterator(self, next_token, max_new_tokens, verbose, print_tokens, cleanup=True):
        output_text, output_text_len = "", 0
        for i in tqdm(range(1, max_new_tokens), disable=(not verbose or print_tokens)):
            next_token = self.gen_next_token(next_token)

            if next_token[0].item() == self.tokenizer.eos_token_id:
                break

            # You need to keep track of the whole text, otherwise you lose spaces, makes everything much slower ¯\_(ツ)_/¯
            # https://github.com/huggingface/transformers/blob/b109257f4fb8b1166e7c53cc5418632014ed53a5/src/transformers/generation/streamers.py#L95-L114
            if print_tokens:
                output_text_len = self.print_current_token(output_text_len)

        torch.cuda.synchronize()
        input_tokens  = self.generated_ids[0, : self.begin_gen_position].cpu()
        output_tokens = self.generated_ids[0, self.begin_gen_position : self.cache_position].cpu()
        output_text   = self.tokenizer.decode(output_tokens)

        if cleanup:
            # model._reset_cache()
            del self.inputs, self.generated_ids, self.cache_position
            torch.cuda.empty_cache()

        return {
            "output_text": output_text,
            "output_tokens": output_tokens,
            "input_tokens": input_tokens,
        }

    @torch.inference_mode()
    def generate(self, prompt, use_chat_template=True, verbose=True, print_tokens=False):
        self.setup(
            inputs=self.tokenize_prompt(prompt, use_chat_template=use_chat_template),
            max_new_tokens=self.max_new_tokens,
        )
        return self.next_token_iterator(self.prefill(), self.max_new_tokens, verbose, print_tokens)

    def generate_(self, prompt, use_chat_template=True, verbose=False, print_tokens=False):
        gen_out = self.model.generate(
            **self.tokenize_prompt(prompt, use_chat_template=use_chat_template),
            do_sample=self.do_sample,
            cache_implementation="static",
            max_new_tokens=self.max_new_tokens,
            pad_token_id=self.tokenizer.pad_token_id,
            temperature=self.temperature,
            top_p=self.top_k,
            # use_cache=False,
        )[0]

        return {"output_text": self.tokenizer.decode(gen_out), "output_tokens": gen_out}

    def tokenize_prompt(self, prompt, use_chat_template=True):
        if use_chat_template:

            prompt = self.tokenizer.apply_chat_template(
                [
                    {"role": "user", "content": prompt},
                ],
                tokenize=False,
                add_generation_prompt=True,
            )

        return self.tokenizer([prompt], return_tensors="pt").to(device=self.model.device)



# SPDX-License-Identifier: Apache-2.0
# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2024

import torch
from torch import Tensor
from ..core.quantize import Quantizer, HQQLinear, BaseQuantizeConfig
from ..core.utils import cleanup
from ..core.peft import HQQLinearLoRA
from ..models.hf.base import AutoHQQHFModel
from ..backends.torchao import patch_hqq_to_aoint4
from termcolor import colored
try:
    from ..backends.marlin import patch_hqq_to_marlin
except Exception:
        patch_hqq_to_marlin = None
try:
    from ..backends.bitblas import patch_hqq_to_bitblas
except Exception:
    patch_hqq_to_bitblas = None

try:
    from ..backends.gemlite import patch_hqq_to_gemlite
except Exception:
    patch_hqq_to_gemlite = None

def patch_linearlayers(model, fct, patch_param=None, verbose=False):

    def _patch_linear(model):
        for name, layer in model.named_children():
            if isinstance(layer, (HQQLinear, HQQLinearLoRA)):
                setattr(model, name, fct(layer, patch_param))
            else:
                _patch_linear(layer)

    _patch_linear(model)


#Automatically build a quant_config if it's not passed as an argument
def autoset_quant_config(hqq_layer, patch_param = None):

    if(patch_param is not None):
        setattr(hqq_layer, 'quant_config', patch_param)

    if(hqq_layer.quant_config is None):
        setattr(hqq_layer, 'quant_config', BaseQuantizeConfig(nbits=hqq_layer.meta['nbits'], 
                                                              group_size=hqq_layer.meta['group_size'], 
                                                              axis=hqq_layer.meta['axis'], 
                                                              quant_scale=hqq_layer.meta['quant_scale'], 
                                                              quant_zero=hqq_layer.meta['quant_zero'],
                                                              ))

    return hqq_layer

def patch_add_quant_config(layer, patch_param):
    if(isinstance(layer, HQQLinear)):
        autoset_quant_config(layer, patch_param)

    if(isinstance(layer, HQQLinearLoRA)):
        autoset_quant_config(layer.linear_layer, patch_param)
    return layer

# add dummy weights to a layer
def patch_add_weight_param(layer, patch_param):
    if hasattr(layer, "weight") is False:
        if hasattr(layer, "device"):
            device_ = layer.device
        else:
            param = [p for p in layer.parameters()]
            device_ = param[0].device if (len(param) > 0) else patch_param["device"]

        fp_param = [p for p in layer.parameters() if p.is_floating_point()]
        dtype_ = fp_param[0].dtype if (len(fp_param) > 0) else patch_param["dtype"]

        layer.weight = torch.nn.Parameter(
            torch.zeros((1,), device=device_, dtype=dtype_), requires_grad=False
        )
    return layer


# Optimize HQQLinear.forward for inference
def patch_hqq_inference(layer, patch_param):
    def forward_hqq_inferece(self, x):
        out = torch.matmul(x.to(self.device), self.dequantize().T)  # TODO GEMV use-case
        if self.bias is not None:
            out += self.bias
        return out

    if type(layer) is HQQLinear:
        layer.forward = lambda x: forward_hqq_inferece(layer, x)

    if type(layer) is HQQLinearLoRA:
        if type(layer.linear_layer) is HQQLinear:
            layer.linear_layer.forward = lambda x: forward_hqq_inferece(
                layer.linear_layer, x
            )

    return layer


# Optimize HQQLinearLoRA.forward for inference
def patch_lora_inference(layer, patch_param):
    def forward_lora_inference(self, x):
        out = torch.matmul(torch.matmul(x, self.lora_A), self.lora_B) * self.scaling
        return out

    if type(layer) is HQQLinearLoRA:
        layer.forward_lora = lambda x: forward_lora_inference(layer, x)
    return layer

# Copied from https://github.com/pytorch/ao/blob/b523f9f9e15b6fb80d10f585d9cf45e0c5e4d10e/torchao/quantization/utils.py#L486-L501
def recommended_inductor_config_setter():
    """
    Set inductor config to use the following optimizations which have been showed to improve performance for quantized models:
        coordinate_descent_tuning = True
        coordinate_descent_check_all_directions = True
        force_fuse_int_mm_with_mul = True
        fx_graph_cache = True
        triton.unique_kernel_names = True
        torch.set_float32_matmul_precision("high")
    """
    torch._inductor.config.coordinate_descent_tuning = True
    torch._inductor.config.coordinate_descent_check_all_directions = True
    torch._inductor.config.force_fuse_int_mm_with_mul = True
    torch._inductor.config.fx_graph_cache = True
    torch._inductor.config.triton.unique_kernel_names = True
    torch.set_float32_matmul_precision("high")


def prepare_for_inference(
    model, allow_merge=False, backend="default", verbose=False
):
    if backend == "torchao_int4":
        allow_merge = False

    patch_linearlayers(model, patch_add_quant_config, patch_param=None)
    patch_linearlayers(model, patch_hqq_inference)
    patch_linearlayers(model, patch_lora_inference)
    cleanup()

    if backend == "gemlite" and (patch_hqq_to_gemlite is not None):
        if patch_hqq_to_gemlite is None:
            raise RunTimeError(
                """
                GemLite backend is not available. Check if gemlite is correctly installed if you want to use the GemLite backend (https://github.com/mobiusml/gemlite/).
                """
            )
        else:
            patch_linearlayers(model, patch_hqq_to_gemlite, verbose=verbose)

    if backend == "bitblas":
        if patch_hqq_to_bitblas is None:
            raise RunTimeError(
                """
                BitBlas backend is not available. Check if bitblas is correctly installed if you want to use the BitBlas backend (https://github.com/mobiusml/bitblas/).
                """
            )
        else:
            patch_linearlayers(model, patch_hqq_to_bitblas, verbose=verbose)

    if backend == "torchao_int4":
        patch_linearlayers(model, patch_hqq_to_aoint4, verbose=verbose)
        recommended_inductor_config_setter()

    if allow_merge:  # only compatible with symmetric quant kernels
        patch_linearlayers(
            model,
            patch_merge_zeros_with_lora,
            {"z_shift": 8, "keep_lora": False},
            verbose=verbose,
        )

    patch_linearlayers(
        model,
        patch_add_weight_param,
        {"device": model.device, "dtype": model.dtype},
    )
    cleanup()


def get_lowrank_tuple_torch_gpu(tensor, max_rank, eps=None):
    t = tensor.t().float()
    u, s, v = torch.linalg.svd(t)
    u, s, v = u[:, :max_rank], s[:max_rank], v[:max_rank, :]
    us = torch.matmul(u, torch.diag(s))
    A, B = (v.t(), us.t())  # t ~ AB
    if eps is not None:
        A = A.clamp(min=-eps, max=eps)
        B = B.clamp(min=-eps, max=eps)
    return A.to(tensor.dtype), B.to(tensor.dtype)


# Merges HQQ's zeros with LoRA weights. ONLY works with group_size=None and axis=1
def patch_merge_zeros_with_lora(layer, patch_params={"z_shift": 8, "keep_lora": False}):
    if type(layer) is HQQLinearLoRA:
        # Check config suppport
        hqq_layer = layer.linear_layer
        if (hqq_layer.meta["axis"] == 0) or (hqq_layer.meta["group_size"] is not None):
            print('Skipping zeros lora merging for', layer.name)
            return layer

        layer.z_shift = patch_params["z_shift"]
        layer.keep_lora = patch_params["keep_lora"]

        shape = layer.linear_layer.meta["shape"]
        z = layer.linear_layer.meta["zero"]
        s = layer.linear_layer.meta["scale"]
        u = (s * (-z + layer.z_shift)).flatten().view([1, -1])

        
        if layer.keep_lora is False:
            A, B = layer.lora_A.data, layer.lora_B.data
            onz = torch.ones((shape[1], 1), device=u.device, dtype=u.dtype)

            # Cat
            A = torch.cat([A, onz], axis=1)
            B = torch.cat([B, u], axis=0)

            # #Re-rank
            # #A, B = get_lowrank_tuple_torch_gpu(torch.matmul(A, B) + torch.matmul(onz, u), max_rank=layer.r + 1)

            layer.lora_A.data = A.to(dtype=layer.lora_A.dtype)
            layer.lora_B.data = B.to(dtype=layer.lora_B.dtype)

            layer.u = None

        else:
            layer.u = torch.nn.Parameter(u, requires_grad=False)
        
        layer.linear_layer.meta["zero"] = 0.0

        torch.cuda.empty_cache()

        def forward_linear_updated(self, x: Tensor) -> Tensor:
            compute_dtype = self.linear_layer.compute_dtype
            meta = self.linear_layer.meta
            W_q = self.linear_layer.W_q

            W_r = Quantizer.unpack[meta["packing"]](W_q, dtype=compute_dtype).t()
            scale = meta["scale"].t()

            out = torch.matmul(x, (W_r - self.z_shift)) * scale  # Symmetric quant
            return out

        layer.linear_layer.forward = lambda x: forward_linear_updated(layer, x)

        def forward_updated(self, x: Tensor) -> Tensor:
            out = self.linear_layer.forward(x)

            if self.u is not None:
                out += torch.matmul(x.sum(axis=-1, keepdim=True), self.u)

            out += self.forward_lora(x) + self.bias
            return out

        layer.forward = lambda x: forward_updated(layer, x)

    return layer


#Loads Peft-HF adapter into an HQQ model
def load_peft_adapter(model, adapter_dir):
    from safetensors import safe_open
    from peft import get_peft_model, PeftConfig

    #Load config
    model = get_peft_model(model, PeftConfig.from_pretrained(adapter_dir))

    #Load weights 
    tensors = {}
    with safe_open(adapter_dir + "/adapter_model.safetensors", framework="pt", device=model.device.type) as f:
        for key in f.keys():
            base, param = '.'.join(key.split('.')[:-1]) + '.default', key.split('.')[-1]
            if(base not in tensors):
                tensors[base] = {}
            tensors[base][param] = torch.nn.Parameter(f.get_tensor(key))

    #Full module name
    for name, module in model.named_modules():
        module.name = name

    #Assign weights
    def _patch(model):
        for name, layer in model.named_children():
            if(layer.name in tensors):
                for p,v in tensors[layer.name].items():
                    setattr(layer, p, v)
            _patch(layer)

    _patch(model)


# SPDX-License-Identifier: Apache-2.0
# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2024

import torch
from ..core.quantize import HQQLinear
from ..core.peft import HQQLinearLoRA

from gemlite.helper import A16Wn

def patch_hqq_to_gemlite(layer, patch_params):
    hqq_layer = None
    if isinstance(layer, HQQLinear):
        hqq_layer = layer
    if isinstance(layer, HQQLinearLoRA):
        hqq_layer = layer.linear_layer

    if hqq_layer is None:
        return layer

    if hqq_layer.meta["group_size"] == None:
        hqq_layer.meta["group_size"] = (
            hqq_layer.in_features if (hqq_layer.meta["axis"] == 1)
            else hqq_layer.out_features
        )

    gemlite_linear = A16Wn(device=hqq_layer.device).from_hqqlinear(hqq_layer)

    if isinstance(layer, HQQLinear):
        return gemlite_linear

    if isinstance(layer, HQQLinearLoRA):
        layer.linear_layer = gemlite_linear

    return layer



# SPDX-License-Identifier: Apache-2.0
# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2024

#re-using some code from vllm Marlin-HQQ, see below.

from typing import Any, Dict, List, Optional
import torch
import logging

from vllm.model_executor.layers.linear import (
    LinearBase,
    LinearMethodBase,
    UnquantizedLinearMethod,
)
from vllm.model_executor.layers.quantization.base_config import (
    QuantizationConfig,
    QuantizeMethodBase,
)

from vllm.model_executor.layers.quantization import register_quantization_config

from vllm.model_executor.layers.quantization.utils.quant_utils import (
    gptq_pack,
    get_pack_factor,
)
from vllm.model_executor.parameter import BasevLLMParameter, PackedvLLMParameter

# HQQ
from vllm.model_executor.layers.quantization.hqq_marlin import (
    HQQMarlinConfig,
    HQQZeroScaleParameter,
    HQQEmptyParameter,
    error_loader,
)
from ..core.quantize import Quantizer, HQQLinear, BaseQuantizeConfig

# Add new linear methods in order to use loader_v2
import vllm.model_executor.layers.linear as vllm_linear

HQQ_LINEAR_METHODS = [
    "HQQGemLiteVLLMLinear",
    "HQQPytorchVLLMLinear",
    "HQQOnTheFlyMethod",
]

for linear_method in HQQ_LINEAR_METHODS:
    if linear_method not in vllm_linear.WEIGHT_LOADER_V2_SUPPORTED:
        vllm_linear.WEIGHT_LOADER_V2_SUPPORTED.append(linear_method)

logger = logging.getLogger(__name__)

# Faster unpacking
def unpack_rows(
    W_q_packed: torch.Tensor,
    W_nbits: int,
    num_output_rows: int,
    num_output_cols: int,
    dtype: torch.dtype = torch.uint8,
) -> torch.Tensor:

    num_rows, num_cols = W_q_packed.shape
    elements_per_sample = num_output_rows // num_rows

    shifts       = torch.arange(elements_per_sample, device=W_q_packed.device, dtype=W_q_packed.dtype) * W_nbits  
    mask         = ((1 << W_nbits) - 1)
    W_q_unpacked = ((W_q_packed.unsqueeze(-1) >> shifts) & mask).to(dtype) 
    W_q_unpacked = W_q_unpacked.permute(0, 2, 1).reshape(num_output_rows, num_cols)
    
    return W_q_unpacked

# Gemlite
try:
    import gemlite
    from gemlite import DType, GemLiteLinear
    from gemlite.core import TORCH_TO_DTYPE

    gemlite_is_available = True

    # Faster gptq_pack
    def gptq_pack(
        q_w: torch.Tensor, num_bits: int, size_k: int, size_n: int
    ) -> torch.Tensor:

        q_w = q_w.cuda()
        out = gemlite.bitpack.pack_weights_over_rows_triton(q_w, num_bits, packing_bitwidth=32, transpose=False)[0]
        del q_w
        torch.cuda.empty_cache()
        return out
        
    # Faster unpacking
    def unpack_rows(
        W_q_packed: torch.Tensor,
        W_nbits: int,
        num_output_rows: int,
        num_output_cols: int,
        dtype: torch.dtype = torch.uint8,
    ) -> torch.Tensor:
        
        return gemlite.bitpack.unpack_over_rows_triton(W_q_packed, W_nbits, num_output_rows, dtype)

except Exception as e:
    logger.warning("Gemlite backend not available. Make sure the lib installed https://github.com/mobiusml/gemlite/", e) 
    gemlite_is_available = False

# Hugging Face config quant name tag
QUANT_NAME = "hqq"


# Override HQQweightParameter to support more nbits.
# TODO: 3-bit support not added yet.
class HQQweightParameter(PackedvLLMParameter):
    def __init__(self, packed_factor: int, packed_dim: int, weight_bits: int, **kwargs):
        super().__init__(packed_factor, packed_dim, None, **kwargs)
        self.weight_bits = weight_bits
        self.packing = Quantizer.bit_to_packing[self.weight_bits]
        self.input_shape = self.shape[self.input_dim] * self.packed_factor
        self.output_shape = self.shape[self.output_dim]

    def load_merged_column_weight(self, loaded_weight: torch.Tensor, **kwargs):
        loaded_weight = Quantizer.unpack[self.packing](loaded_weight, dtype=torch.uint8)
        loaded_weight = loaded_weight.reshape(-1, self.input_shape).transpose(1, 0)
        loaded_weight = gptq_pack(
            loaded_weight,
            self.weight_bits,
            loaded_weight.shape[0],
            loaded_weight.shape[1],
        )
        super().load_merged_column_weight(loaded_weight, **kwargs)

    def load_row_parallel_weight(self, loaded_weight: torch.Tensor):
        loaded_weight = Quantizer.unpack[self.packing](loaded_weight, dtype=torch.uint8)
        loaded_weight = loaded_weight.reshape(self.output_shape, -1).transpose(1, 0)
        loaded_weight = gptq_pack(
            loaded_weight,
            self.weight_bits,
            loaded_weight.shape[0],
            loaded_weight.shape[1],
        )
        super().load_row_parallel_weight(loaded_weight)

    def load_qkv_weight(self, loaded_weight: torch.Tensor, **kwargs):
        loaded_weight = Quantizer.unpack[self.packing](loaded_weight, dtype=torch.uint8)
        loaded_weight = loaded_weight.reshape(-1, self.input_shape).transpose(1, 0)
        loaded_weight = gptq_pack(
            loaded_weight,
            self.weight_bits,
            loaded_weight.shape[0],
            loaded_weight.shape[1],
        )
        super().load_qkv_weight(loaded_weight, **kwargs)


class HQQOnTheFlyweightParameter(PackedvLLMParameter):
    def __init__(self, packed_factor: int, packed_dim: int, weight_bits: int, **kwargs):
        super().__init__(packed_factor, packed_dim, None, **kwargs)
        self.weight_bits = weight_bits
        self.input_shape = self.shape[self.input_dim] * self.packed_factor
        self.output_shape = self.shape[self.output_dim]

    def load_merged_column_weight(self, loaded_weight: torch.Tensor, **kwargs):
        loaded_weight = loaded_weight.view(-1, self.input_shape)
        super().load_merged_column_weight(loaded_weight, **kwargs)

    def load_row_parallel_weight(self, loaded_weight: torch.Tensor):
        loaded_weight = loaded_weight.view(self.output_shape, -1)
        super().load_row_parallel_weight(loaded_weight)

    def load_qkv_weight(self, loaded_weight: torch.Tensor, **kwargs):
        loaded_weight = loaded_weight.view(-1, self.input_shape)
        super().load_qkv_weight(loaded_weight, **kwargs)


#Create alias
@register_quantization_config("hqq_marlin")
class HQQMarlinPrequantizedConfig(HQQMarlinConfig):
    @classmethod
    def get_name(cls) -> str:
        return 'hqq_marlin'

##############################################################################################
#HQQBaseVLLMConfig / HQQBaseVLLMLinear
##############################################################################################
# Base HQQ/VLLM Linear method
class HQQBaseVLLMConfig(QuantizationConfig):
    """Config class for HQQ"""

    def __init__(
        self,
        weight_bits: int,
        group_size: int,
        axis: int = 1,
        skip_modules: Optional[List[str]] = None,
    ) -> None:

        self.weight_bits = weight_bits
        self.group_size = group_size
        self.pack_factor = 32 // weight_bits  # pre-packed into int32 in GPTQ format
        self.skip_modules = skip_modules
        self.packing = Quantizer.bit_to_packing[self.weight_bits]
        self.axis = axis

    def __repr__(self) -> str:
        return (
            f"HQQBaseVLLMConfig(weight_bits={self.weight_bits}, "
            f"group_size={self.group_size})"
        )

    @classmethod
    def get_name(cls) -> str:
        return QUANT_NAME

    @classmethod
    def get_supported_act_dtypes(cls) -> List[torch.dtype]:
        return [torch.float16, torch.bfloat16]

    @classmethod
    def get_min_capability(cls) -> int:
        return 75

    @classmethod
    def get_config_filenames(cls) -> List[str]:
        return ["quantize_config.json"]

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "HQQBaseVLLMConfig":
        wq_params = config["quant_config"]["weight_quant_params"]
        weight_bits = cls.get_from_keys(wq_params, ["nbits"])
        axis = cls.get_from_keys(wq_params, ["axis"])
        group_size = cls.get_from_keys(wq_params, ["group_size"])
        skip_modules = config["skip_modules"]
        return cls(weight_bits, group_size, axis, skip_modules)

    def is_layer_skipped(self, prefix: str) -> bool:
        # Split the prefix into its dot-separated components
        components = prefix.split(".")

        # Check if any of the skip modules exactly matches any component
        return self.skip_modules is not None and any(
            module_name in components for module_name in self.skip_modules
        )

    def get_quant_method(
        self, layer: torch.nn.Module, prefix: str
    ) -> Optional["QuantizeMethodBase"]:
        if isinstance(layer, LinearBase):
            if self.is_layer_skipped(prefix):
                return UnquantizedLinearMethod()
            return HQQPytorchVLLMLinear(self)
        return None

class HQQBaseVLLMLinear(LinearMethodBase):
    """Base HQQ Linear VLLM method"""

    def __init__(self, quant_config: QuantizationConfig):
        self.quant_config = quant_config

    def create_weights(
        self,
        layer: torch.nn.Module,
        input_size_per_partition: int,
        output_partition_sizes: List[int],
        input_size: int,
        output_size: int,
        params_dtype: torch.dtype,
        **extra_weight_attrs,
    ) -> None:

        self.output_size_per_partition = sum(output_partition_sizes)
        self.input_size_per_partition = input_size_per_partition

        weight_loader = extra_weight_attrs.get("weight_loader", error_loader)

        self.scales_and_zp_size = (
            input_size_per_partition // self.quant_config.group_size
        )

        # Transposed - GPTQ/GemLited packed
        W_q = HQQweightParameter(
            data=torch.empty(
                self.input_size_per_partition // self.quant_config.pack_factor,
                self.output_size_per_partition,
                dtype=torch.int32,
            ),
            input_dim=0,
            output_dim=1,
            packed_dim=0,
            packed_factor=self.quant_config.pack_factor,
            weight_bits=self.quant_config.weight_bits,
            weight_loader=weight_loader,
        )
        layer.register_parameter("W_q", W_q)

        zero = HQQZeroScaleParameter(
            data=torch.empty(
                self.output_size_per_partition,
                self.scales_and_zp_size,
                dtype=params_dtype,
            ),
            input_dim=1,
            output_dim=0,
            weight_loader=weight_loader,
        )
        layer.register_parameter("zero", zero)

        scale = HQQZeroScaleParameter(
            data=torch.empty(
                self.output_size_per_partition,
                self.scales_and_zp_size,
                dtype=params_dtype,
            ),
            input_dim=1,
            output_dim=0,
            weight_loader=weight_loader,
        )
        layer.register_parameter("scale", scale)

        setattr(
            layer,
            "current_shape",
            [self.output_size_per_partition, self.input_size_per_partition],
        )

        layer.register_parameter(
            "shape",
            BasevLLMParameter(
                data=torch.empty(2, dtype=torch.int), weight_loader=weight_loader
            ),
        )
        layer.register_parameter(
            "nbits",
            BasevLLMParameter(
                data=torch.empty(1, dtype=torch.int), weight_loader=weight_loader
            ),
        )

        ignore_parameters = (
            "axis",
            "channel_wise",
            "compute_dtype",
            "encoded_state_dict",
            "group_size",
            "offload_meta",
            "optimize",
            "packing",
            "quant_scale",
            "quant_zero",
            "round_zero",
            "stores_quant_config",
            "unpack_view_dtype",
            "view_as_float",
        )

        for name in ignore_parameters:
            layer.register_parameter(
                name,
                HQQEmptyParameter(data=torch.empty(0), weight_loader=weight_loader),
            )

    def unpack(self, layer, dtype):
        return unpack_rows(
            layer.W_q,
            self.quant_config.weight_bits,
            self.input_size_per_partition,
            self.output_size_per_partition,
            dtype=dtype,
        )

    def process_weights_after_loading(self, layer: torch.nn.Module) -> None:
        # Shape
        setattr(layer, "orig_shape", torch.Size(layer.shape))
        del layer.shape

        setattr(layer, "W_nbits", int(layer.nbits))
        del layer.nbits

        del layer.compute_dtype
        setattr(layer, "compute_dtype", layer.scale.dtype)


##############################################################################################
#HQQPytorchConfig / HQQPytorchVLLMLinear
##############################################################################################
# Pytorch
class HQQPytorchConfig(HQQBaseVLLMConfig):
    """Config class for HQQ"""

    def __init__(
        self,
        weight_bits: int,
        group_size: int,
        axis: int = 1,
        skip_modules: Optional[List[str]] = None,
    ) -> None:
        super().__init__(weight_bits, group_size, axis, skip_modules)

    @classmethod
    def get_supported_act_dtypes(cls) -> List[torch.dtype]:
        return [torch.float16, torch.bfloat16]

    def get_quant_method(
        self, layer: torch.nn.Module, prefix: str
    ) -> Optional["QuantizeMethodBase"]:
        if isinstance(layer, LinearBase):
            if self.is_layer_skipped(prefix):
                return UnquantizedLinearMethod()
            return HQQPytorchVLLMLinear(self)
        return None

#Create alias
@register_quantization_config("hqq_torch")
class HQQPytorchPrequantizedConfig(HQQPytorchConfig):
    @classmethod
    def get_name(cls) -> str:
        return 'hqq_torch'

class HQQPytorchVLLMLinear(HQQBaseVLLMLinear):
    """Linear HQQ VLLM with Pytorch backend"""

    def __init__(self, quant_config: QuantizationConfig):
        super().__init__(quant_config)

    def process_weights_after_loading(self, layer: torch.nn.Module) -> None:
        super().process_weights_after_loading(layer)

        # Repack for faster dequant
        group_size = self.quant_config.group_size
        W_q = (
            self.unpack(layer, dtype=layer.compute_dtype)
            .T.reshape(-1, group_size)
            .contiguous()
        )

        layer.W_q   = torch.nn.Parameter(Quantizer.pack[self.quant_config.packing](W_q), requires_grad=False)
        layer.scale = torch.nn.Parameter(layer.scale.data, requires_grad=False)
        layer.zero  = torch.nn.Parameter(layer.zero.data, requires_grad=False)
        torch.cuda.empty_cache()

    @torch.compile()
    def dequantize(self, layer):  # Only 8, 4, 2, 1 bit support. 3-bit NOT supported yet
        scale = layer.scale.view(-1, 1)  # non-transposed
        zero = layer.zero.view(-1, 1)  # non-transposed

        group_size = self.quant_config.group_size
        W_q = Quantizer.unpack[self.quant_config.packing](layer.W_q, dtype=layer.compute_dtype)
 
        if(self.quant_config.weight_bits == 3):
            W_q = W_q[: group_size if self.axis == 0 else self.current_shape[0] * self.current_shape[1] // group_size]

        W_q = W_q.view(-1, group_size)
        W_r = ((W_q - zero) * scale).view(layer.current_shape)
        return W_r

    def apply(
        self,
        layer: torch.nn.Module,
        x: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        W_r = self.dequantize(layer).T
        out = torch.matmul(x, W_r)

        if bias is not None:
            out += bias

        return out


##############################################################################################
#HQQGemLiteConfig / HQQGemLiteVLLMLinear
##############################################################################################
# GemLite
class HQQGemLiteConfig(HQQBaseVLLMConfig):
    """Config class for HQQ"""

    def __init__(
        self,
        weight_bits: int,
        group_size: int,
        axis: int = 1,
        skip_modules: Optional[List[str]] = None,
    ) -> None:
        assert axis == 1, 'Only axis=1 is supported.'
        super().__init__(weight_bits, group_size, 1, skip_modules)

    @classmethod
    def get_supported_act_dtypes(cls) -> List[torch.dtype]:
        return [torch.float16, torch.bfloat16]

    def get_quant_method(
        self, layer: torch.nn.Module, prefix: str
    ) -> Optional["QuantizeMethodBase"]:
        if isinstance(layer, LinearBase):
            if self.is_layer_skipped(prefix):
                return UnquantizedLinearMethod()
            try:
                return HQQGemLiteVLLMLinear(self)
            except:
                return HQQPytorchVLLMLinear(self)
        return None

#Create alias
@register_quantization_config("hqq_gemlite")
class HQQGemLitePrequantizedConfig(HQQGemLiteConfig):
    @classmethod
    def get_name(cls) -> str:
        return 'hqq_gemlite'

class HQQGemLiteVLLMLinear(HQQBaseVLLMLinear):
    """Linear HQQ VLLM with GemLite backend"""

    def __init__(
        self,
        quant_config: QuantizationConfig,
    ):
        super().__init__(quant_config)

    def process_weights_after_loading(self, layer: torch.nn.Module) -> None:
        super().process_weights_after_loading(layer)

        # Gemlite
        dtype = layer.scale.dtype
        gemlite_dtype = TORCH_TO_DTYPE[dtype]
        gemlite_linear = GemLiteLinear(
            self.quant_config.weight_bits,
            group_size=self.quant_config.group_size,
            in_features=self.input_size_per_partition,
            out_features=self.output_size_per_partition,
            input_dtype=gemlite_dtype,
            output_dtype=gemlite_dtype,
            scaled_activations=False,
        )

        gemlite_linear.pack(
            self.unpack(layer, dtype=torch.uint8)
            .T.contiguous()
            .view(layer.current_shape),
            layer.scale.view(-1, 1),
            layer.zero.view(-1, 1),
            bias=None,
        )

        layer.gemlite_linear = gemlite_linear
        del layer.W_q, layer.scale, layer.zero

        torch.cuda.empty_cache()

    def apply(
        self,
        layer: torch.nn.Module,
        x: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        out = layer.gemlite_linear(x)

        if bias is not None:
            out += bias

        return out

##############################################################################################
#VLLM_HQQ_BACKEND
##############################################################################################
# Backends
class VLLM_HQQ_BACKEND:
    MARLIN = HQQMarlinConfig
    GEMLITE = HQQGemLiteConfig
    PYTORCH = HQQPytorchConfig


DEFAULT_VLLM_HQQ_BACKEND = VLLM_HQQ_BACKEND.GEMLITE
COMPILE_OPTIONS = {}  # {"mode":"max-autotune-no-cudagraphs"}


##############################################################################################
#HQQOnTheFlyConfig
##############################################################################################
# On-the-fly quantization
@register_quantization_config("hqq_onthefly")
class HQQOnTheFlyConfig(QuantizationConfig):
    """Config class for HQQ Marlin"""

    def __init__(
        self,
        weight_bits: int,
        group_size: int,
        quant_mode: str = "static",
        skip_modules: Optional[List[str]] = None,
    ) -> None:
        self.weight_bits = weight_bits
        self.group_size = group_size
        self.quant_mode = quant_mode
        self.skip_modules = skip_modules

    def __repr__(self) -> str:
        return (
            f"HQQOnTheFlyConfig(weight_bits={self.weight_bits}, "
            f"group_size={self.group_size})"
        )

    @classmethod
    def get_name(cls) -> str:
        return "hqq_onthefly"

    @classmethod
    def get_supported_act_dtypes(cls) -> List[torch.dtype]:
        return [torch.float16, torch.bfloat16]

    @classmethod
    def get_min_capability(cls) -> int:
        return 75

    @classmethod
    def get_config_filenames(cls) -> List[str]:
        return ["quantize_config.json"]

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "HQQOnTheFlyConfig":
        wq_params = config["quant_config"]["weight_quant_params"]
        weight_bits = cls.get_from_keys(wq_params, ["nbits"])
        quant_mode = cls.get_from_keys(wq_params, ["quant_mode"])
        group_size = cls.get_from_keys(wq_params, ["group_size"])
        skip_modules = config["skip_modules"]

        return cls(weight_bits, group_size, quant_mode, skip_modules)

    def is_layer_skipped(self, prefix: str) -> bool:
        # Split the prefix into its dot-separated components
        components = prefix.split(".")

        # Check if any of the skip modules exactly matches any component
        return self.skip_modules is not None and any(
            module_name in components for module_name in self.skip_modules
        )

    def get_quant_method(
        self, layer: torch.nn.Module, prefix: str
    ) -> Optional["QuantizeMethodBase"]:
        if isinstance(layer, LinearBase):
            if self.is_layer_skipped(prefix):
                return UnquantizedLinearMethod()
            return HQQOnTheFlyMethod(self)
        return None


class HQQOnTheFlyMethod(LinearMethodBase):
    """Linear HQQ"""

    enable_compile = True  # compile the forward pass

    def __init__(self, quant_config: QuantizationConfig):
        self.quant_config = quant_config

    def create_weights(
        self,
        layer: torch.nn.Module,
        input_size_per_partition: int,
        output_partition_sizes: List[int],
        input_size: int,
        output_size: int,
        params_dtype: torch.dtype,
        **extra_weight_attrs,
    ) -> None:
        self.output_size_per_partition = sum(output_partition_sizes)
        self.input_size_per_partition = input_size_per_partition
        self.params_dtype = params_dtype

        weight_loader = extra_weight_attrs.get("weight_loader", error_loader)
        
        weight = HQQOnTheFlyweightParameter(
            data=torch.empty(
                self.output_size_per_partition,
                self.input_size_per_partition,
                dtype=params_dtype,
            ),
            input_dim=1,
            output_dim=0,
            packed_dim=0,
            packed_factor=1,
            weight_bits=16,
            weight_loader=weight_loader,
        )

        layer.register_parameter("weight", weight)

    def process_weights_after_loading(self, layer: torch.nn.Module) -> None:
        global DEFAULT_VLLM_HQQ_BACKEND

        device     = layer.weight.device
        W_nbits    = self.quant_config.weight_bits
        group_size = self.quant_config.group_size
        quant_mode = self.quant_config.quant_mode
        shape      = layer.weight.shape
        tmp_linear = None

        # Marlin
        if DEFAULT_VLLM_HQQ_BACKEND == VLLM_HQQ_BACKEND.MARLIN:
            raise NotImplementedError("MARLIN backend is not implemented yet for on-the-fly quantization.")

        processor_args = {'device': device, 'dtype':self.params_dtype}

        if W_nbits == 8:
            layer.quant_layer = None

            if DEFAULT_VLLM_HQQ_BACKEND == VLLM_HQQ_BACKEND.GEMLITE:
                #8-bit weights
                tmp_linear = torch.nn.Linear(1, 1, bias=False)
                tmp_linear.weight.data = layer.weight
                tmp_linear.in_features = layer.weight.shape[1]
                tmp_linear.out_features = layer.weight.shape[0]
        
                if "dynamic" in quant_mode:
                    processor = gemlite.helper.A8W8_int8_dynamic #default: A8W8 int8
                    if 'fp8' in quant_mode:
                        processor = gemlite.helper.A8W8_fp8_dynamic #A8W8 fp8
                    if 'mxfp8' in quant_mode:
                        processor_args['post_scale'] = True if (group_size is None) else False #MXPF8 x MXPF8
                        processor = gemlite.helper.A8W8_MXFP_dynamic
                else:
                    if('mxfp8' in quant_mode): #A16W8 - weight-only
                        processor = gemlite.helper.A16W8_MXFP
                    else:
                        processor = gemlite.helper.A16W8 
                try:
                    layer.quant_layer = processor(**processor_args).from_linear(tmp_linear)
                except Exception:
                    logger.error(f'Failed gemlite conversion, using Pytorch backend. weights.shape={shape}, processor:{processor}') 

            if DEFAULT_VLLM_HQQ_BACKEND == VLLM_HQQ_BACKEND.PYTORCH or (layer.quant_layer is None):
                layer.quant_layer = HQQLinear.from_weights(
                    layer.weight,
                    bias=None,
                    quant_config=BaseQuantizeConfig(nbits=W_nbits, group_size=group_size, axis=1),
                    compute_dtype=self.params_dtype,
                    device=device,
                )

        elif W_nbits == 4:
            # Quantized weights
            if('mxfp' in quant_mode or 'nvfp' in quant_mode):
                if DEFAULT_VLLM_HQQ_BACKEND != VLLM_HQQ_BACKEND.GEMLITE:
                    raise NotImplementedError('Unsupported backend for MXFP/NVFP')

                tmp_linear = torch.nn.Linear(1, 1, bias=False)
                tmp_linear.weight.data = layer.weight
                tmp_linear.in_features = layer.weight.shape[1]
                tmp_linear.out_features = layer.weight.shape[0]
                layer.quant_layer = tmp_linear

                if('mxfp8' in quant_mode):
                    processor = gemlite.helper.A8W4_MXFP_dynamic #MXFP8 x MXFP4
                if('mxfp4' in quant_mode):
                    if('dynamic' in quant_mode): #MXFP4 x MXPF4
                        processor = gemlite.helper.A4W4_MXFP_dynamic
                    else:
                        processor = gemlite.helper.A16W4_MXFP #MXFP4 weight-only
                elif('nvfp4' in quant_mode):
                    processor = gemlite.helper.A4W4_NVFP_dynamic

            else: #INT mode vvia HQQ
                layer.weight = layer.weight.to(device).contiguous()
                layer.quant_layer = HQQLinear.from_weights(
                    layer.weight,
                    bias=None,
                    quant_config=BaseQuantizeConfig(nbits=W_nbits, group_size=group_size, axis=1),
                    compute_dtype=self.params_dtype,
                    device=device,
                )
                processor = gemlite.helper.A16Wn

            if DEFAULT_VLLM_HQQ_BACKEND == VLLM_HQQ_BACKEND.GEMLITE:
                try:
                    if(isinstance(layer.quant_layer, HQQLinear)):
                        layer.quant_layer = processor(**processor_args).from_hqqlinear(layer.quant_layer)
                    else:
                        layer.quant_layer = processor(**processor_args).from_linear(layer.quant_layer)
                except Exception:
                    logger.error(f'Failed A16Wn gemlite conversion, using Pytorch backend. weights.shape={shape}') 

        #Clean-up
        torch.cuda.empty_cache()

        # Compile
        if HQQOnTheFlyMethod.enable_compile:
            layer.quant_layer.forward = torch.compile(layer.quant_layer.forward, **COMPILE_OPTIONS)

        # Cleanup
        del layer.weight, tmp_linear
        torch.cuda.empty_cache()

    def apply(
        self,
        layer: torch.nn.Module,
        x: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        out = layer.quant_layer(x)

        if bias is not None:
            out += bias

        return out


##############################################################################################
#Main patching functions
##############################################################################################
# Allows overriding a VLLM quant method with arbitrary configs
def enable_vllm_hqq_quant():
    return 
    
def patch_vllm_quant_method(quant_name: str, quant_config: QuantizationConfig):
    import vllm.model_executor.layers.quantization as vllm_quantization

    get_quantization_config_orig = vllm_quantization.get_quantization_config

    def get_quantization_config_patched(quantization: str):
        if quantization == quant_name:
            return quant_config
        else:
            return get_quantization_config_orig(quantization)

    vllm_quantization.get_quantization_config = get_quantization_config_patched


def set_vllm_hqq_backend(backend: QuantizationConfig):
    global DEFAULT_VLLM_HQQ_BACKEND
    DEFAULT_VLLM_HQQ_BACKEND = backend
    if (not gemlite_is_available) and (backend == VLLM_HQQ_BACKEND.GEMLITE):
        logger.error(
            "The GemLite backend is not availble. Make sure gemlite is installed: https://github.com/mobiusml/gemlite"
        )
    return patch_vllm_quant_method(QUANT_NAME, backend)


# Set's on-the-fly hqq quantization for vllm
def set_vllm_onthefly_hqq_quant(
    weight_bits=4, group_size=64, quant_mode="static", skip_modules=["lm_head", "visual", "vision"]
):
    from vllm.model_executor.layers import linear

    original_init = linear.LinearBase.__init__

    def patched_init(
        self,
        input_size,
        output_size,
        skip_bias_add=False,
        params_dtype=None,
        quant_config=None,
        *args,
        **kwargs,
    ):
        if quant_config is None:
            quant_config = HQQOnTheFlyConfig(
                weight_bits, group_size, quant_mode, skip_modules
            )
        original_init(
            self,
            input_size,
            output_size,
            skip_bias_add,
            params_dtype,
            quant_config,
            *args,
            **kwargs,
        )

    linear.LinearBase.__init__ = patched_init


##############################################################################################
# Model specific patching
##############################################################################################

def patch_mixtral():
    import torch
    import torch.nn as nn
    from typing import Optional
    from vllm.model_executor.layers.linear import RowParallelLinear
    from vllm.model_executor.layers.quantization.base_config import QuantizationConfig
    import vllm.model_executor.models.mixtral_quant as mixtral_quant

    # Mixtral
    class MixtralMLPRowParallel(nn.Module):
        def __init__(
            self,
            num_experts: int,
            hidden_size: int,
            intermediate_size: int,
            quant_config: Optional[QuantizationConfig] = None,
        ) -> None:
            super().__init__()
            self.num_experts = num_experts
            self.ffn_dim = intermediate_size
            self.hidden_dim = hidden_size

            self.w1 = RowParallelLinear(
                self.hidden_dim, self.ffn_dim, bias=False, quant_config=quant_config
            )
            self.w2 = RowParallelLinear(
                self.ffn_dim, self.hidden_dim, bias=False, quant_config=quant_config
            )
            self.w3 = RowParallelLinear(
                self.hidden_dim, self.ffn_dim, bias=False, quant_config=quant_config
            )

            # TODO: Use vllm's SiluAndMul
            self.act_fn = nn.SiLU()

        def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
            w1_out, _ = self.w1(hidden_states)
            w1_out = self.act_fn(w1_out)
            w3_out, _ = self.w3(hidden_states)
            current_hidden_states = w1_out * w3_out
            current_hidden_states, _ = self.w2(current_hidden_states)
            return current_hidden_states

    mixtral_quant.MixtralMLP = MixtralMLPRowParallel
